// This is an MPI+CUDA program that does the following:
//
// 1. On the host, fill the A and B arrays with random numbers
// 2. On the host, print the initial values of the A and B arrays
// 3. Copy the A and B arrays from the host to the device
// 4. On the device, add the A and B vectors and store the result in C
// 5. Copy the C array from the device to the host
// 6. On the host, print the result
//
// Author: Aaron Weeden, Shodor, 2016

// Import library so we can call MPI functions
#include <mpi.h>

// Import library so we can call printf()
#include <stdio.h>

// Import library so we can call exit(), malloc(), free(), random(), etc.
#include <stdlib.h>

// Import library so we can call time()
#include <time.h>

// Import library with model parameters
#include "params.h"

// Import library with declarations for functions that use CUDA
#include "add-vectors.h"

// Declare variables for the host and device arrays
int * HostA;
int * HostB;
int * HostC;

// Declare functions that will be defined later
void TryMPI(int const err);
void TryMalloc(void * const err);

// Start the program
int main(int argc, char ** argv)
{
  // Initialize MPI, set rank and size
  MPI_Init(&argc, &argv);
  int rank;
  TryMPI(MPI_Comm_rank(MPI_COMM_WORLD, &rank));
  int size;
  TryMPI(MPI_Comm_size(MPI_COMM_WORLD, &size));

  // Make it so we can catch the return code of MPI functions
  MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_RETURN);

  // Allocate memory for the host arrays
  TryMalloc(HostA = (int*)malloc(BYTE_COUNT));
  TryMalloc(HostB = (int*)malloc(BYTE_COUNT));
  TryMalloc(HostC = (int*)malloc(BYTE_COUNT));

  // Allocate memory for the device arrays
  InitDeviceMemory();

  // Initialize the random number generator
  srandom(time(NULL));

  // On the host, fill the A and B arrays with random numbers
  if (rank == 0)
  {
    printf("Expected Result:\n");
  }
  for (int i = 0; i < NUM_COUNT; i++)
  {
    HostA[i] = 100 * random() / RAND_MAX;
    HostB[i] = 100 * random() / RAND_MAX;
    if (rank == 0)
    {
      printf("\tHostC[%d] should be %d + %d\n", i, HostA[i], HostB[i]);
    }
  }

  // 1. Copy the A and B arrays from the host to the device
  // 2. On the device, add the A and B vectors and store the result in C
  // 3. Copy the C array from the device to the host
  AddVectorsOnDevice();

  int allCs[size * NUM_COUNT];
  TryMPI(MPI_Gather(HostC, // Where in memory is the message that will be 
                           // sent?
                    NUM_COUNT, // How many elements of the message will be sent
                               // from each process?
                    MPI_INT, // What is the datatype of the elements to be
                             // sent?
                    allCs, // Where in memory should the received message 
                           // be stored?
                    NUM_COUNT, // How many elements of the message will be sent
                               // from each process?
                    MPI_INT, // What is the datatype of the elements to be
                             // received?
                    0, // Who is the receiver?
                    MPI_COMM_WORLD)); // Which processes are involved in this 
                                      // communication?

  // On the host, print the result
  if (rank == 0)
  {
    printf("Result:\n");
    for (int i = 0; i < size * NUM_COUNT; i++)
    {
      printf("\tHostC[%d] = %d\n", i, allCs[i]);
    }
  }

  // De-allocate memory for the device arrays
  FinalizeDeviceMemory();

  // De-allocate memory for the host arrays
  free(HostC);
  free(HostB);
  free(HostA);

  // Finish calling MPI functions
  TryMPI(MPI_Finalize());

  return 0;
}

// Checks if the return to an MPI function is MPI_SUCCESS, and exits if it is
// not
void TryMPI(int const err)
{
  if (err != MPI_SUCCESS)
  {
    char string[120];
    int resultlen;
    MPI_Error_string(err, string, &resultlen);
    fprintf(stderr, "ERROR: MPI: %s\n", string);
    MPI_Abort(MPI_COMM_WORLD, err);
  }
}

// Define a function to check whether a malloc() call was successful
void TryMalloc(void * const err)
{
  if (err == NULL)
  {
    fprintf(stderr, "malloc error\n");
    exit(EXIT_FAILURE);
  }
}

